#from mlxtend.evaluate import permutation_test
from statsmodels.tsa import stattools
from statsmodels.stats import diagnostic
from statsmodels.tsa import arima_model
import warnings
from collections import namedtuple
from sklearn.metrics import mean_squared_error
from sklearn.model_selection import train_test_split
import numpy as np
from sklearn import metrics
from scipy.stats import mstats
warnings.filterwarnings("ignore")

class TimeSeriesMetrics:
    def __init__(self):
        pass
    
    @staticmethod
    def unscaled_mean_bounded_relative_absolute_error(y_true, y_pred):
        """
        Unscaled Mean Bounded Relative Absolute Error
        Formula taken from:
        https://www.ncbi.nlm.nih.gov/pmc/articles/PMC5365136/
        @y_true - Y[i]
        @y_pred - F[i]
        """
        numerator = [abs(elem - y_pred[idx]) for idx, elem in enumerate(y_true)]
        series_one = y_true[1:]
        series_two = y_true[:-1]
        denominator = [abs(elem - series_two[idx]) for idx, elem in enumerate(series_one)]
        final_series = [numerator[idx]/(numerator[idx] + denominator[idx])
                        for idx in range(len(denominator))]
        mbrae = np.mean(final_series)
        return mbrae/(1-mbrae)

    @staticmethod
    def mean_bounded_relative_absolute_error(y_true, y_pred):
        """
        Mean Bounded Relative Absolute Error
        Formula taken from:
        https://www.ncbi.nlm.nih.gov/pmc/articles/PMC5365136/
        @y_true - Y[i]
        @y_pred - F[i]
        """
        numerator = [abs(elem - y_pred[idx]) for idx, elem in enumerate(y_true)]
        series_one = y_true[1:]
        series_two = y_true[:-1]
        denominator = [abs(elem - series_two[idx]) for idx, elem in enumerate(series_one)]
        final_series = [numerator[idx]/(numerator[idx] + denominator[idx])
                        for idx in range(len(denominator))]
        return np.mean(final_series)

    @staticmethod
    def mean_absolute_percentage_error(y_true, y_pred): 
        y_true, y_pred = np.array(y_true), np.array(y_pred)
        return np.mean(np.abs((y_true - y_pred) / y_true)) * 100

    @staticmethod
    def root_mean_squared_error(y_true, y_pred):
        return np.sqrt(metrics.mean_squared_error(y_true, y_pred))

    @staticmethod
    def mean_relative_absolute_error(y_true, y_pred):
        """
        formula comes from: 
        http://www.spiderfinancial.com/support/documentation/numxl/reference-manual/forecasting-performance/mrae
        """
        numerator = [abs(elem - y_pred[idx])
                     for idx, elem in enumerate(y_true)]
        series_one = y_true[1:]
        series_two = y_true[:-1]
        denominator = [abs(elem - series_two[idx])
                       for idx, elem in enumerate(series_one)]    
        return np.mean([
            numerator[i]/denominator[i] for i in range(len(numerator))])

    @staticmethod
    def median_relative_absolute_error(y_true, y_pred):
        """
        formula comes from: 
        http://www.spiderfinancial.com/support/documentation/numxl/reference-manual/forecasting-performance/mrae
        """
        numerator = [abs(elem - y_pred[idx])
                     for idx, elem in enumerate(y_true)]
        series_one = y_true[1:]
        series_two = y_true[:-1]
        denominator = [abs(elem - series_two[idx])
                       for idx, elem in enumerate(series_one)]    
        return np.median([
            numerator[i]/denominator[i] for i in range(len(numerator))])

    @staticmethod
    def symmetric_mean_absolute_percentage_error(y_true, y_pred):
        """
        formula comes from:
        https://en.wikipedia.org/wiki/Symmetric_mean_absolute_percentage_error
        """
        numerator = [abs(y_pred[idx] - elem) for idx, elem in enumerate(y_true)]
        denominator = [abs(elem) + abs(y_pred[idx]) for idx, elem in enumerate(y_true)]
        denominator = [elem/2 for elem in denominator]
        result = np.mean([numerator[i]/denominator[i] for i in range(len(numerator))])
        return result * 100

    @staticmethod
    def symmetric_median_absolute_percentage_error(y_true, y_pred):
        """
        formula comes from:
        https://en.wikipedia.org/wiki/Symmetric_mean_absolute_percentage_error
        """
        numerator = [abs(y_pred[idx] - elem) for idx, elem in enumerate(y_true)]
        denominator = [abs(elem) + abs(y_pred[idx]) for idx, elem in enumerate(y_true)]
        denominator = [elem/2 for elem in denominator]
        result = np.median([numerator[i]/denominator[i] for i in range(len(numerator))])
        return result * 100

    @staticmethod
    def mean_absolute_scaled_error(y_true, y_pred):
        """
        formula comes from:
        https://en.wikipedia.org/wiki/Mean_absolute_scaled_error
        """
        numerator = sum([abs(y_pred[idx] - elem)  for idx, elem in enumerate(y_true)])
        series_one = y_true[1:]
        series_two = y_true[:-1]
        denominator = sum([abs(elem - series_two[idx])
                       for idx, elem in enumerate(series_one)])
        coeficient = len(y_true)/(len(y_true)-1)
        return numerator/(coeficient * denominator)

    @staticmethod
    def geometric_mean_relative_absolute_error(y_true, y_pred):
        numerator = [abs(y_pred[idx] - elem)  for idx, elem in enumerate(y_true)]
        series_one = y_true[1:]
        series_two = y_true[:-1]
        denominator = [abs(elem - series_two[idx])
                       for idx, elem in enumerate(series_one)]
        return mstats.gmean([numerator[i]/denominator[i] for i in range(len(numerator))])


class TimeSeriesHypothesisTests:
    def __init__(self, timeseries):
        pass

    @staticmethod
    def ad_fuller_test(self, timeseries):
        result = stattools.adfuller(timeseries)
        AdFullerResult = namedtuple('AdFullerResult', 'statistic pvalue')
        return AdFullerResult(result[0], result[1])

    @staticmethod
    def kpss(self, timeseries):
        result = stattools.kpss(timeseries)
        KPSSResult = namedtuple('KPSSResult', 'statistic pvalue')
        return KPSSResult(result[0], result[1])

    @staticmethod
    def cointegration(self, timeseries, alt_timeseries):
        result = stattools.coint(timeseries, alt_timeseries)
        CointegrationResult = namedtuple('CointegrationResult', 'statistic pvalue')
        return CointegrationResult(result[0], result[1])

    @staticmethod
    def bds(self):
        result = stattools.bds(self.timeseries)
        BdsResult = namedtuple('BdsResult', 'statistic pvalue')
        return BdsResult(result[0], result[1])

    @staticmethod
    def q_stat(self, timeseries):
        autocorrelation_coefs = stattools.acf(timeseries)
        result = stattools.q_stat(autocorrelation_coefs)
        QstatResult = namedtuple('QstatResult', 'statistic pvalue')
        return QstatResult(result[0], result[1])

    def _evaluate_arima_model(self, X, arima_order):
        # prepare training dataset
        train, test, _, _ = train_test_split(X, np.zeros(X.shape[0]))
        history = list(train)
        # make predictions
        predictions = list()
        for t in range(len(test)):
            model = ARIMA(history, order=arima_order)
            model_fit = model.fit(disp=0)
            yhat = model_fit.forecast()[0]
            predictions.append(yhat)
            history.append(test[t])
        # calculate out of sample error
        error = mean_squared_error(test, predictions)
        return error

    # evaluate combinations of p, d and q values for an ARIMA model
    @staticmethod
    def generate_model(self, timeseries):
        best_score, best_cfg = float("inf"), None
        p_values = [0, 1, 2, 4, 6, 8, 10]
        d_values = range(0, 3)
        q_values = range(0, 3)
        for p in p_values:
            for d in d_values:
                for q in q_values:
                    order = (p,d,q)
                    try:
                        mse = self._evaluate_arima_model(timeseries, order)
                        if mse < best_score:
                            best_score = mse
                            best_order = order
                    except:
                        continue
        model = ARIMA(timeseries, order=best_order)
        model_result = model.fit(disp=0)
        return model, model_result

    @staticmethod
    def acorr_ljungbox(self, timeseries):
        model, model_result = self.generate_model(timeseries)
        result = diagnostic.acorr_ljungbox(model_result)
        AcorrLjungBoxResult = namedtuple('AcorrLjungBoxResult', 'statistic pvalue')
        return AcorrLjungBoxResult(result[0], result[1])

    @staticmethod
    def acorr_breusch_godfrey(self, timeseries):
        model, model_result = self.generate_model(timeseries)
        result = diagnostic.acorr_breusch_godfrey(model_result)
        AcorrBreuschGodfreyResult = namedtuple('BreuschGodfreyResult', 'statistic pvalue')
        return AcorrBreuschGodfreyResult(result[0], result[1])

    @staticmethod
    def het_arch(self, timeseries):
        model, model_result = self.generate_model(timeseries)
        result = diagnostic.het_arch(model_result)
        HetArchResult = namedtuple('HetArchResult', 'statistic pvalue')
        return HetArchResult(result[0], result[1])

    @staticmethod
    def breaks_cumsumolsresid(self, timeseries):
        model, model_result = self.generate_model(timeseries)
        result = diagnostic.breaks_cusumolsresid(model_result)
        BreaksCumSumResult = namedtuple('BreaksCumSumResult', 'statistic pvalue')
        return BreaksCumSumResult(result[0], result[1])




from describer_ml.numeric.num_stats import trimean

def get_diffs(listing):
    diffs = []
    for elem in listing:
        for other in listing:
            diffs.append(
                abs(elem - other)
            )
    return diffs

def within_tolerance(diffs, max_diff):
    for diff in diffs:
        if diff > max_diff:
            return False
    return True

def get_matches(grouped_df, max_diff):
    matching_columns = []
    for column in grouped_df.columns:
        diffs = get_diffs(grouped_df[column])
        if within_tolerance(diffs, max_diff):
            matching_columns.append(column)
    return matching_columns

def mean_match(df, match_column, max_diff):
    average_per_class = df.groupby(match_column).mean()
    return get_matches(mean_per_class, max_diff)

def median_match(df, match_column, max_diff):
    median_per_class = df.groupby(match_column).median()
    return get_matches(median_per_class, max_diff)

def trimean_match(df, match_column, max_diff):
    trimean_per_class = df.groupby(match_column).agg(trimean)
    return get_matches(trimean_per_class, max_diff)



import pandas as pd
import numpy as np
from thinkbayes2 import Cdf as CDF
from scipy import stats
from scipy.spatial import distance
from sklearn.ensemble import IsolationForest
from sklearn import preprocessing 

def minimum(array):
    return np.amin(array)

def maximum(array):
    return np.amax(array)

def minimum_with_nan(array):
    return np.nanmin(array)

def maximum_with_nan(array):
    return np.nanmax(array)

def percentile(array):
    return np.percentile(array)

def percentile_with_nan(array):
    return np.nanpercentile(array)

def quantile(array):
    return np.quantile(array)

def quantile_with_nan(array):
    return np.nanquantile(array)

def median(array):
    return np.median(array)

def mean(array):
    return np.mean(array)

def standard_deviation(array):
    return np.std(array)

def variance(array):
    return np.variance(array)

def median_with_nan(array):
    return np.nanmedian(array)

def mean_with_nan(array):
    return np.nanmean(array)

def standard_deviation_with_nan(array):
    return np.nanstd(array)

def variance_with_nan(array):
    return np.nanvar(array)

def geometric_mean(array):
    return stats.gmean(array)

def harmonic_mean(array):
    return stats.hmean(array)

def kurtosis(array):
    return stats.kurtosis(array)

def mode(array):
    return stats.mode(array)

def skew(array):
    return stats.skew(array)

def variation(array):
    return stats.variation(array)

def find_repeats(array):
    return stats.find_repeats(array)

def interquartile_range(array):
    return stats.iqr(array)

def entropy(probabilities, alternative_probabilities=None):
    return stats.entropy(probabilities, alternative_probabilities)

def trimean(data):
    q1 = np.quantile(data, 0.25)
    q3 = np.quantile(data, 0.75)
    median = np.median(data)
    return (q1 + 2*median + q3)/4

def interquartile_mean(data):
    q1 = np.quantile(data, 0.25)
    q3 = np.quantile(data, 0.75)
    sorted_data = np.sort(data)
    trimmed_data = sorted_data[(sorted_data >= q1) & (sorted_data <= q3)]
    return np.mean(trimmed_data)

def midhinge(data):
    q1 = np.quantile(data, 0.25)
    q3 = np.quantile(data, 0.75)
    return np.mean([q1, q3])

def value_range(data):
    max_val = np.max(data)
    min_val = np.min(data)
    return abs(max_val - min_val)

def trimean_absolute_deviation(data):
    trimean = trimean(data)
    numerator = [abs(elem - trimean) for elem in data]
    return sum(numerator)/len(data)

def mean_absolute_deviation(data):
    mean = mean(data)
    numerator = [abs(elem - mean) for elem in data]
    return sum(numerator)/len(data)

def median_absolute_deviation(data):
    median = median(data)
    numerator = [abs(elem - median) for elem in data]
    return sum(numerator)/len(data)

def _get_cdf(dist):
    cdf = CDF(dist)
    return dict(
        zip(
            list(cdf.xs),
            list(cdf.ps)
        )
    )

def _get_prob_values(cdf):
    return list(cdf.values())

def get_inliers_outliers_if(dist):
    clf = IsolationForest(
        behaviour='new', contamination='auto'
    )
    result = clf.fit_predict(dist.reshape(-1, 1))
    result = pd.Series(result)
    outliers = dist[result[result == -1].index]
    inliers = dist[result[result != -1].index]
    return inliers, outliers
    
def isclose(value_one, value_two,
            max_deviance,
            distance_function=None):
    """
    Here we define the max_deviance parameter.
    This parameter assumes you have a sense of
    the appropriate deviance.
    I suggest looking at some plots of your data,
    when doing exploratory data analysis to figure this out.
    The appropriate deviance is a maximum threshold for what
    should be considered close in your dataset.
    Eventually, the goal is to figure out an automated method
    to figure this out.  However, for now, no such metric
    appears obvious, therefore I leave it to the user to figure out
    what 'closeness' means for your data.
    
    If this measure is hard to come by, or changes often, then this
    is probably not the right metric for you.
    
    Some sense of stability in the scale of your data is expected,
    to make use of this function.
    """
    if not distance_function:
        distance_function = distance.euclidean
    if distance_function(value_one, value_two) < max_deviance:
        return True
    else:
        return False

def get_compare_value(value, cdf,
                      max_deviance,
                      distance_function=None):
    """
    Please see doc string in isclose.
    """
    if value in cdf:
        return value
    for value_two in cdf:
        if np.isclose(value, value_two):
            return value_two
        close_enough = isclose(
            value, value_two, max_deviance,
            distance_function=distance_function
        )
        if close_enough:
            return value_two
    return None

def get_within_boundary(cdf_one, cdf_two, spread,
                        max_deviance,
                        distance_function=None):
    """
    Please see doc string in isclose. 
    """
    within_upper_bound = []
    within_lower_bound = []
    for value in cdf_one:
        other_value = get_compare_value(
            value, cdf_two, max_deviance,
            distance_function=distance_function
        )
        if not other_value:
            within_upper_bound.append(False)
            within_lower_bound.append(False)
        else:
            within_upper_bound.append(
                cdf_two[other_value] < cdf_one[value] + spread
            )
            within_lower_bound.append(
                cdf_two[other_value] > cdf_one[value] - spread
            )
    within_upper_bound = np.array(within_upper_bound)
    within_lower_bound = np.array(within_lower_bound)
    return within_upper_bound & within_lower_bound

def get_inliers(dist):
    dist = np.array(dist)
    inliers, outliers = get_inliers_outliers_if(dist)
    return inliers

def compare_cdf_mean_absolute_deviation(dist_one, dist_two,
                                        max_deviance,
                                        distance_function=None,
                                        remove_outliers=True):
    """
    We assume dist_one and dist_two are of the same size.
    I.E. len(dist_one) == len(dist_two)
    
    Please see doc string for isclose.
    """
    if remove_outliers:
        dist_one = get_inliers(dist_one)
        dist_two = get_inliers(dist_two)
    cdf_one = _get_cdf(dist_one)
    cdf_two = _get_cdf(dist_two)
    mad = mean_absolute_deviation(_get_prob_values(cdf_one))
    within_boundary = get_within_boundary(
        cdf_one, cdf_two, mad,
        max_deviance,
        distance_function=distance_function
    )
    return (within_boundary).sum()/len(dist_one)

def compare_cdf_median_absolute_deviation(dist_one, dist_two,
                                          max_deviance,
                                          distance_function=None,
                                          remove_outliers=True):
    """
    We assume dist_one and dist_two are of the same size.
    I.E. len(dist_one) == len(dist_two)

    Please see doc string for isclose
    """
    if remove_outliers:
        dist_one = get_inliers(dist_one)
        dist_two = get_inliers(dist_two)
    cdf_one = _get_cdf(dist_one)
    cdf_two = _get_cdf(dist_two)
    mad = median_absolute_deviation(_get_prob_values(cdf_one))
    within_boundary = get_within_boundary(
        cdf_one, cdf_two, mad,
        max_deviance,
        distance_function=distance_function
    )
    return (within_boundary).sum()/len(dist_one)

def compare_cdf_trimean_absolute_deviation(dist_one, dist_two,
                                           max_deviance,
                                           distance_function=None,
                                           remove_outliers=True):
    """
    We assume dist_one and dist_two are of the same size.
    I.E. len(dist_one) == len(dist_two)

    Please see doc string for isclose
    """
    if remove_outliers:
        dist_one = get_inliers(dist_one)
        dist_two = get_inliers(dist_two)
    cdf_one = _get_cdf(dist_one)
    cdf_two = _get_cdf(dist_two)
    tad = trimean_absolute_deviation(_get_prob_values(cdf_one))
    within_boundary = get_within_boundary(
        cdf_one, cdf_two, tad,
        max_deviance,
        distance_function=distance_function
    )
    return (within_boundary).sum()/len(dist_one)

def compare_cdf_hard_coded_boundary(dist_one, dist_two,
                                    max_deviance,
                                    distance_function=None,
                                    boundary=0.01,
                                    remove_outliers=True):
    """
    We assume dist_one and dist_two are of the same size.
    I.E. len(dist_one) == len(dist_two)

    Please see doc string for isclose
    """
    if remove_outliers:
        dist_one = get_inliers(dist_one)
        dist_two = get_inliers(dist_two)
    cdf_one = _get_cdf(dist_one)
    cdf_two = _get_cdf(dist_two)
    within_boundary = get_within_boundary(
        cdf_one, cdf_two, boundary,
        max_deviance,
        distance_function=distance_function
    )
    return (within_boundary).sum()/len(dist_one)

# things like this
# trimmed statistics
# investigate here: https://en.wikipedia.org/wiki/Descriptive_statistics



import pathlib
from setuptools import setup

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

# This call to setup() does all the work
setup(
    name="describer_ml",
    version="0.15",
    description="A set of descriptive statistics and hypothesis tests",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/EricSchles/describer_ml",
    author="Eric Schles",
    author_email="ericschles@gmail.com",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
    packages=["describer_ml", "describer_ml.timeseries", "describer_ml.numeric", "describer_ml.matching"],
    include_package_data=True,
    install_requires=["sklearn", "scipy", "numpy", "statsmodels", "pytest", "mlxtend"],
)
